package main

import (
	"fmt"
	"io"
	"os"
	"strings"

	"github.com/alexflint/go-arg"
	"go.wit.com/lib/gui/shell"
	"go.wit.com/log"
)

// sent via -ldflags
var VERSION string
var BUILDTIME string

func main() {
	pp := arg.MustParse(&argv)

	// you need a proto file
	if argv.Proto == "" {
		log.Info("you must provide --proto <filename>")
		os.Exit(-1)
	}

	if !shell.Exists(argv.Proto) {
		log.Info("protobuf", argv.Proto, "is missing")
		if !argv.DryRun {
			os.Exit(-1)
		}
	}

	if !strings.HasSuffix(argv.Proto, ".proto") {
		log.Info("protobuf", argv.Proto, "must end in .proto")
		os.Exit(-1)
	}

	// you need --upbase and --lobase
	if argv.UpBase == "" {
		pp.WriteHelp(os.Stdout)
		os.Exit(-1)
	}

	cmd := []string{"go", "list", "-f", "'{{.Name}}'"}
	result := shell.Run(cmd)

	packageName := strings.Join(result.Stdout, "\n")
	packageName = strings.TrimSpace(packageName)
	packageName = strings.Trim(packageName, "'")
	log.Info("packageName == ", packageName)

	protobase := strings.TrimSuffix(argv.Proto, ".proto")

	sortmap := make(map[string]string)
	sortmap["package"] = packageName
	sortmap["protobase"] = protobase
	sortmap["base"] = argv.LoBase
	sortmap["lock"] = sortmap["base"] + "slock"
	sortmap["Base"] = argv.UpBase
	sortmap["Bases"] = sortmap["Base"] + "s"

	sortparts := strings.Split(argv.Sort[0], ",")
	sortmap["sortBy"] = sortparts[0]
	sortmap["sortKey"] = sortparts[1]

	if argv.DryRun {
		for k, v := range sortmap {
			log.Info(k, "=", v)
		}
		os.Exit(0)
	}

	f, _ := os.OpenFile(protobase+".sort.pb.go", os.O_WRONLY|os.O_CREATE, 0600)

	header(f, sortmap)
	syncLock(f, sortmap)
	iterTop(f, sortmap)
	iterNext(f, sortmap)
	iterSort(f, sortmap)
	iterAppend(f, sortmap)
	iterEnd(f, sortmap)

	// make the foo.marshal.pb.go file
	marshal(sortmap)
}

func headerComment(w io.Writer) {
	fmt.Fprintln(w, "")
	fmt.Fprintln(w, "// this file was autogenerated with autogenpb")
	fmt.Fprintln(w, "//")
	fmt.Fprintln(w, "// you might be able to use it on simple, strictly defined protobuf files")
	fmt.Fprintln(w, "//")
	fmt.Fprintln(w, "// go install go.wit.com/apps/autogenpb@latest")
	fmt.Fprintln(w, "")
}

func header(w io.Writer, names map[string]string) {
	fmt.Fprintln(w, "package "+names["package"])
	headerComment(w)
	fmt.Fprintln(w, "import (")
	fmt.Fprintln(w, "	\"fmt\"")
	fmt.Fprintln(w, "	\"os\"")
	fmt.Fprintln(w, "	\"sort\"")
	fmt.Fprintln(w, "	\"sync\"")
	fmt.Fprintln(w, ")")
	fmt.Fprintln(w, "")
}

func syncLock(w io.Writer, names map[string]string) {
	fmt.Fprintln(w, "// bad global lock until I figure out some other plan")
	fmt.Fprintln(w, "var "+names["lock"]+" sync.RWMutex")
	fmt.Fprintln(w, "")
}

func iterTop(w io.Writer, names map[string]string) {
	fmt.Fprintln(w, "type "+names["Base"]+"Iterator struct {")
	fmt.Fprintln(w, "	sync.RWMutex")
	fmt.Fprintln(w, "")
	fmt.Fprintln(w, "	packs []*"+names["Base"])
	fmt.Fprintln(w, "	index int")
	fmt.Fprintln(w, "}")
	fmt.Fprintln(w, "")
	fmt.Fprintln(w, "// New"+names["Base"]+"Iterator initializes a new iterator.")
	fmt.Fprintln(w, "func New"+names["Base"]+"Iterator(packs []*"+names["Base"]+") *"+names["Base"]+"Iterator {")
	fmt.Fprintln(w, "	return &"+names["Base"]+"Iterator{packs: packs}")
	fmt.Fprintln(w, "}")
	fmt.Fprintln(w, "")
	fmt.Fprintln(w, "// Scan moves to the next element and returns false if there are no more packs.")
	fmt.Fprintln(w, "// Use Scan() in a loop, similar to a while loop")
	fmt.Fprintln(w, "//")
	fmt.Fprintln(w, "//	for iterator.Scan() ")
	fmt.Fprintln(w, "//		d := iterator.Next(")
	fmt.Fprintln(w, "//		fmt.Println(\"found UUID:\", d.Uuid")
	fmt.Fprintln(w, "//	}")
	fmt.Fprintln(w, "func (it *"+names["Base"]+"Iterator) Scan() bool {")
	fmt.Fprintln(w, "	if it.index >= len(it.packs) {")
	fmt.Fprintln(w, "		return false")
	fmt.Fprintln(w, "	}")
	fmt.Fprintln(w, "	it.index++")
	fmt.Fprintln(w, "	return true")
	fmt.Fprintln(w, "}")
	fmt.Fprintln(w, "")
}

func iterNext(w io.Writer, names map[string]string) {
	fmt.Fprintln(w, "// Next() returns the next thing in the array")
	fmt.Fprintln(w, "func (it *"+names["Base"]+"Iterator) Next() *"+names["Base"]+" {")
	fmt.Fprintln(w, "	if it.packs[it.index-1] == nil {")
	fmt.Fprintln(w, "		for i, d := range it.packs {")
	fmt.Fprintln(w, "			fmt.Println(\"i =\", i, d)")
	fmt.Fprintln(w, "		}")
	fmt.Fprintln(w, "		fmt.Println(\"protobuf autogenpb sort error len =\", len(it.packs))")
	fmt.Fprintln(w, "		fmt.Println(\"protobuf autogenpb sort error next == nil\", it.index, it.index-1)")
	fmt.Fprintln(w, "		os.Exit(-1)")
	fmt.Fprintln(w, "	}")
	fmt.Fprintln(w, "	return it.packs[it.index-1]")
	fmt.Fprintln(w, "}")
	fmt.Fprintln(w, "")
}

func iterSort(w io.Writer, names map[string]string) {
	fmt.Fprintln(w, "func (all *"+names["Bases"]+") All() *"+names["Base"]+"Iterator {")
	fmt.Fprintln(w, "	"+names["base"]+"Pointers := all.selectAll"+names["Base"]+"()")
	fmt.Fprintln(w, "")
	fmt.Fprintln(w, "	iterator := New"+names["Base"]+"Iterator("+names["base"]+"Pointers)")
	fmt.Fprintln(w, "	return iterator")
	fmt.Fprintln(w, "}")
	fmt.Fprintln(w, "")
	fmt.Fprintln(w, "func (all *"+names["Bases"]+") Sort"+names["sortBy"]+"() *"+names["Base"]+"Iterator {")
	fmt.Fprintln(w, "	packs := all.selectAll"+names["Base"]+"()")
	fmt.Fprintln(w, "")
	fmt.Fprintln(w, "	sort.Sort("+names["Base"]+""+names["sortBy"]+"(packs))")
	fmt.Fprintln(w, "")
	fmt.Fprintln(w, "	iterator := New"+names["Base"]+"Iterator(packs)")
	fmt.Fprintln(w, "	return iterator")
	fmt.Fprintln(w, "}")
	fmt.Fprintln(w, "")
	fmt.Fprintln(w, "func (all *"+names["Bases"]+") Len() int {")
	fmt.Fprintln(w, "	"+names["lock"]+".RLock()")
	fmt.Fprintln(w, "	defer "+names["lock"]+".RUnlock()")
	fmt.Fprintln(w, "")
	fmt.Fprintln(w, "	return len(all."+names["Bases"]+")")
	fmt.Fprintln(w, "}")
	fmt.Fprintln(w, "")
}

func iterEnd(w io.Writer, names map[string]string) {
	fmt.Fprintln(w, "type "+names["Base"]+""+names["sortBy"]+" []*"+names["Base"]+"")
	fmt.Fprintln(w, "")
	fmt.Fprintln(w, "func (a "+names["Base"]+""+names["sortBy"]+") Len() int           { return len(a) }")
	fmt.Fprintln(w, "func (a "+names["Base"]+""+names["sortBy"]+") Less(i, j int) bool { return a[i]."+names["sortKey"]+" < a[j]."+names["sortKey"]+" }")
	fmt.Fprintln(w, "func (a "+names["Base"]+""+names["sortBy"]+") Swap(i, j int)      { a[i], a[j] = a[j], a[i] }")
	fmt.Fprintln(w, "")
	fmt.Fprintln(w, "// safely returns a slice of pointers to the "+names["Base"]+" protobufs")
	fmt.Fprintln(w, "func (all *"+names["Bases"]+") selectAll"+names["Base"]+"() []*"+names["Base"]+" {")
	fmt.Fprintln(w, "	"+names["lock"]+".RLock()")
	fmt.Fprintln(w, "	defer "+names["lock"]+".RUnlock()")
	fmt.Fprintln(w, "")
	fmt.Fprintln(w, "	// Create a new slice to hold pointers to each "+names["Base"]+"")
	fmt.Fprintln(w, "	var aStuff []*"+names["Base"]+"")
	fmt.Fprintln(w, "	aStuff = make([]*"+names["Base"]+", len(all."+names["Bases"]+"))")
	fmt.Fprintln(w, "	for i, p := range all."+names["Bases"]+" {")
	fmt.Fprintln(w, "		aStuff[i] = p // Copy pointers for safe iteration")
	fmt.Fprintln(w, "	}")
	fmt.Fprintln(w, "")
	fmt.Fprintln(w, "	return aStuff")
	fmt.Fprintln(w, "}")
	fmt.Fprintln(w, "")
}

func iterAppend(w io.Writer, names map[string]string) {
	fmt.Fprintln(w, "// enforces no duplicate Refname names")
	fmt.Fprintln(w, "func (all *"+names["Bases"]+") Append(newP *"+names["Base"]+") bool {")
	fmt.Fprintln(w, "	"+names["lock"]+".Lock()")
	fmt.Fprintln(w, "	defer "+names["lock"]+".Unlock()")
	fmt.Fprintln(w, "")
	fmt.Fprintln(w, "	for _, p := range all."+names["Bases"]+" {")
	fmt.Fprintln(w, "		if p.Refname == newP.Refname {")
	fmt.Fprintln(w, "			return false")
	fmt.Fprintln(w, "		}")
	fmt.Fprintln(w, "	}")
	fmt.Fprintln(w, "")
	fmt.Fprintln(w, "	all."+names["Bases"]+" = append(all."+names["Bases"]+", newP)")
	fmt.Fprintln(w, "	return true")
	fmt.Fprintln(w, "}")
	fmt.Fprintln(w, "")
}
